! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!error
! Module: implot_mod
!
! Authors:  Clive Page - University of Leicester
!	    Extensively hacked by Ian Stewart from Clive's implot-1.6
!          
! Description: Plots EPIC/OM data using PGPLOT and WCS
! 
! Comments:
!	TODO:
!	- Implement \param{removeblack}.
!	- Re labels and radii: make the failure kinder if the requisite columns are not there. Ie check to see first, and eg if the radius column is not present, warn then make radii constant.
!	- Add 'constant' as a value for --radiusstyle.
!	- Don't attempt to plot sources for which the --radiusexpression is null. Warn of such if any occur.
!	- some of the readColDataToFixed could be replaced by readColDataToPtr.
!	- pgmtxt after pgwedg: read unit from image for this.
!
! implot-1.6   Version with consistent internal/external numbering!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module implot_mod

  use dal
  use cal
  use param  
  use errorHandling
  use metatask_mod ! to get task name and version number.
  use datafilter

  use constants
  use wcs_aux
  use ssc_misc		! for intToStr, fracVal, ucase
  use array_utils	! for valueAtHistoFraction, median
  use dal_aux		! for maxNonNullValue, readArrayData, splitSetTabName
  use ellipse
  use cal_aux
  use sort_mod
  use source_cutouts
use dump

  implicit none 

  private 

  ! Command-line parameters:
  character(255) ::&
    pgplotDevice = '',&
    imageSetName = '',&
    srcListSetName = '',&
    srcListSetTabName = '',&
    selectionExpression = '',&
    maxSourceNumSortExpression = '',&
    srcLabelExpression = '',&
    radiusExpression = '',&
    tempSrcSetName = ''
  character(20) ::&
    srcListTabName = '',&
    radiusStyle = '',&
    maxSourceNumSortStyle = '',&
    srcLabelStyle = '',&
    gridStyle = '',&
    imageStyle = '',&
    imageTransferStyle = '',&
    contourStyle = ''
  character(20), allocatable ::&
    userLabels(:)
  integer ::&
    colourMapId = 0,&
    textClrIndex = 0,&
    lineThicknessPixels = 0,&
    numContours = 0
  integer, allocatable ::&
    userSrcClrIndices(:)
  integer(int32) ::&
    userMaxSourceNum = 0
  real ::&
    maxRadiusPixels = 0.0,&
    userRadiusPixels = 0.0,&
    textSize = 0.0
  real(single) ::&
    zClipFractionLo = 0.0,&
    zClipFractionHi = 0.0,&
    fracBackground = 0.0,&
    xClipFractionLo = 0.0,&
    xClipFractionHi = 0.0,&
    yClipFractionLo = 0.0,&
    yClipFractionHi = 0.0,&
    contourSeparation = 0.0
  logical ::&
    withMaxSrcNum = .false.,&
    withSrcLabels = .false.,&
    withZclip = .false.,&
    withSrcList = .false.,&
    clipBlankBorder = .false.,&
    withEllipse = .false.,&
    withXYclip = .false.,&
    addUserSources = .false.,&
    withFrame = .false.
!  type(srcDataType), allocatable ::&
!    userSrcData(:)
  real(double), allocatable ::&
    userRaDeg(:), userDecDeg(:)

  ! Other globals:
  type(WcsT) ::&
    wcs
  character(len=8) ::&
    wcsType
  integer(int32) ::&
    instrumentId
  type(SpacecraftAttitudeType) ::&
    scAttitude, instrumAttitude
  real(double) ::&
    expStartTimeStamp,&
    pixelAreaArcsec
  integer ::&
    imageTransferId
  logical ::&
    deviceIsPrinter

  real, parameter ::&
    mainVpXMin = 0.19,&
    mainVpXMax = 0.81,&
    mainVpYMin = 0.156,&
    mainVpYMax = 0.95,&
    mainCharSize = 0.7,&
    rhVpXMin = mainVpXMax,&
    rhVpXMax = 1.0,&
    rhVpYMin = 0.206,&
    rhVpYMax = 1.0,&
    rhCharSize = 0.5,&
    rhCharX = 0.05,&
    rhCharDy = 0.05,&
    rhCharXindent = 0.04,&
    rhCharYindent = -0.02,&
    botVpXMin = 0.1,&
    botVpXMax = 1.0,&
    botVpYMin = 0.0,&
    botVpYMax = 0.1,&
    botCharSize = 0.7,&
    botCharX = 0.0,&
    botCharDy = 0.21
!    botCharXindent = 0.04,&
!    botCharYindent = -0.02

logical :: testFlag = .false.

  public implot

contains 

  !---------------------------------------------------------------------
  subroutine implot
  !---------------------------------------------------------------------

    implicit none

    integer :: ierr, pgbeg
    real(single), pointer :: image(:,:)
    real :: effMin, effMax
    integer(int16) :: minx, maxx, miny, maxy ! image pixel ranges
!    logical(bool), allocatable :: nonNullsArray(:,:)
    logical, allocatable :: nonNullsArray(:,:)
    character(3) :: pgDeviceRump
    integer :: i, slashPosition

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    call message(AppMsg, Verbose, 'Reading image.')
    call readArrayData(imageSetName, image) ! in ssclib/dal_aux

if (gridStyle .ne. 'none' .or. withSrcList .or. addUserSources) then
!*** what about when image is a DETXY image?? Is there any way to handle this??
  call readWcs(imageSetName, 0_int32, wcs)
  call easifyWcsType(wcs%X%ctype, wcs%Y%ctype, wcsType)
end if

    allocate(nonNullsArray(size(image, 1), size(image, 2)))
    call fixUpNulls(image, nonNullsArray)

    if (withZclip) then
      effMin = valueAtHistoFraction(real(image), zClipFractionLo, nonNullsArray)
      effMax = valueAtHistoFraction(real(image), zClipFractionHi, nonNullsArray)
!      effMin = valueAtHistoFraction(real(image), zClipFractionLo, logical(nonNullsArray))
!      effMax = valueAtHistoFraction(real(image), zClipFractionHi, logical(nonNullsArray))
!**** the mask argument is logical not logical(bool) since ssclib-3.17.
      ! function valueAtHistoFraction() is in ssclib/array_utils.
    else
      effMin = minval(image)
      effMax = maxval(image)
      ! Don't need to mask with nonNullsArray - recall that fixUpNulls()
      ! replaces all nulls in the image by the minimum non-null value.
    end if

if (effMin > effMax) call error('badZclipLevels','The minimum and &
&maximum image brightness truncation levels are inverted! This probably &
&means there is a bug somewhere. Please contact the code developer.')

    ! set up PGPLOT.  Note use PGBEG not PGOPEN because latter
    ! does not get picked up from library on some installations
    ierr = pgbeg(0, trim(pgplotDevice), 1, 1)
    if(ierr /= 1)&
call error("badPGPLOTdevice", "Invalid PGPLOT device: "//trim(pgplotDevice))

    ! Avoid irritating phenomenon in which background and foreground
    ! colours swap for ps/cps output, except within the image:
    slashPosition = index(pgplotDevice, "/")
    pgDeviceRump = ucase(pgplotDevice(slashPosition+1:))
    if ((pgDeviceRump .eq. 'PS' .or. pgDeviceRump .eq. 'CPS')&
    .and. imageStyle .eq. 'image') then
      deviceIsPrinter = .true.

      if (textClrIndex .eq. 1) then
        textClrIndex = 0 
      else if (textClrIndex .eq. 0) then
        textClrIndex = 1
      end if

      do i = 1, parameterCount('userras')
        if(userSrcClrIndices(i) .eq. 1) then
          userSrcClrIndices(i) = 0
        else if(userSrcClrIndices(i) .eq. 0) then
          userSrcClrIndices(i) = 1
        end if
      end do
    end if ! device is postscript.

    call plotImage(image, effMin, effMax, minx, maxx, miny, maxy)
    pixelAreaArcsec = wcsPixelAreaWorld(wcs) * 3600.0 * 3600.0
!*** should really check that worldUnits are 'deg'

    if(withSrcList) call addSources(minx, maxx, miny, maxy)
    call pgend
    deallocate(image)

  end subroutine implot

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    integer :: numUserSrcs, i

    !-------------------------------------------------------------------
    ! Start

    imageSetName = stringParameter('set')
    pgplotDevice = stringParameter('device')
    gridStyle = stringParameter('gridstyle')

    imageStyle = stringParameter('imagestyle')
    select case(imageStyle)
    case('image')
      imageTransferStyle = stringParameter('zscaletype')
      select case(imageTransferStyle)
      case('linear')
        imageTransferId = 0
      case('log')
        imageTransferId = 1
      case('sqrt')
        imageTransferId = 2
      case default
call error('badImageTransferId','The image transfer id (parameter &
&--zscaletype) is not recognized.')
      end select

      colourMapId = intParameter('colourmap')

    case('contour')
      contourStyle = stringParameter('contourstyle')
      numContours = intParameter('ncontours')
      contourSeparation = realParameter('contourseparation')

    case default
call error('badImageStyle','The image style (parameter &
&--imagestyle) is not recognized.')
    end select

    withZclip = booleanParameter('withzclip')
    if (withZclip) then
      zClipFractionLo = realParameter('zhistolo')
      zClipFractionHi = realParameter('zhistohi')

      if (zClipFractionLo .ge. zClipFractionHi)&
call error('badZclipValues','The requested --zhistolo is greater than &
&or equal to --zhistohi.')
    end if

    clipBlankBorder = booleanParameter('trimborder')
    withFrame = booleanParameter('withframe')

    withXYclip = booleanParameter('withxyclip')
    if (withXYclip) then
      xClipFractionLo = realParameter('xfraclo')
      xClipFractionHi = realParameter('xfrachi')
      yClipFractionLo = realParameter('yfraclo')
      yClipFractionHi = realParameter('yfrachi')
    end if

    withSrcList = booleanParameter('withsrclisttab')
    if (withSrcList) then
      srcListSetTabName = stringParameter('srclisttab')
      selectionExpression = stringParameter('expression')
      lineThicknessPixels = intParameter('thickness')
      textClrIndex = intParameter('srccolour')
      tempSrcSetName = stringParameter('tempset')

      withMaxSrcNum = booleanParameter('withncut')
      if (withMaxSrcNum) then
        userMaxSourceNum = intParameter('ncut')

        maxSourceNumSortStyle = stringParameter('ncutsortstyle')
        select case(maxSourceNumSortStyle)
        case('expr')
          maxSourceNumSortExpression = stringParameter('ncutsortexpression')

        case('radiusup')

        case('radiusdown')

        case('rownumber')

        case default
call error('badNcutSortStyle','The max source determination style (parameter &
&--ncutsortstyle) is not recognized.')
        end select
      end if

      maxRadiusPixels = realParameter('maxradius')

      radiusStyle = stringParameter('radiusstyle')
      select case(radiusStyle)
      case('user')
        userRadiusPixels = realParameter('userradius')

      case('expr')
        radiusExpression = stringParameter('radiusexpression')

      case('psf')
        withEllipse = booleanParameter('withellipse')
        fracBackground = realParameter('bkgfraction')

      case default
call error('badRadiusStyle','The source radius style (parameter &
&--radiusstyle) is not recognized.')
      end select

      withSrcLabels = booleanParameter('withlabels')
      if (withSrcLabels) then
        textSize = realParameter('labelsize')

        srcLabelStyle = stringParameter('labelstyle')
        select case(srcLabelStyle)
        case('expr')
          srcLabelExpression = stringParameter('labelexpression')

        case('radiusup')

        case('radiusdown')

        case('rownumber')

        case default
call error('badLabelStyle','The source label style (parameter &
&--labelstyle) is not recognized.')
        end select
      end if

      addUserSources = booleanParameter('addusersources')
      if (addUserSources) then
        numUserSrcs = parameterCount('userras')
        allocate(userRaDeg(numUserSrcs), userDecDeg(numUserSrcs)&
          , userLabels(numUserSrcs), userSrcClrIndices(numUserSrcs))
        do i = 0, numUserSrcs - 1
!          userSrcData(i + 1)%raDeg  = realParameter('userras',      i)
!          userSrcData(i + 1)%decDeg = realParameter('userdecs',     i)
          userRaDeg(i + 1)          = realParameter('userras',      i)
          userDecDeg(i + 1)         = realParameter('userdecs',     i)
          userLabels(i + 1)         = stringParameter('userlabels', i)
          userSrcClrIndices(i + 1)  = intParameter('usercolours',   i)
        end do
      end if
    end if

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine fixUpNulls(image, nonNullsArray)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(inout) :: image(:,:)
!    logical(bool), intent(out) :: nonNullsArray(:,:)
    logical,      intent(out)   :: nonNullsArray(:,:)

    real :: amin, amax

    !-------------------------------------------------------------------
    ! Start

    amin = minval(image)          ! seems to ignore null values
    amax = maxval(image) 

    ! It seems that no comparative returns T for NaNs, so I constructed
    ! this double-negative test that seems to work:

    nonNullsArray = .true. ! default
    where(.not.(image .ge. amin) .and. .not.(image .le. amax))
      nonNullsArray = .false.
      image = amin
    end where

  end subroutine fixUpNulls

  !---------------------------------------------------------------------
  subroutine plotImage(image, effMin, effMax, minx, maxx, miny, maxy)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: image(:,:)
    real, intent(in) :: effMin, effMax   ! effective range of image
    integer(int16), intent(out) :: minx, maxx, miny, maxy

    integer :: status, xsize, ysize
    real :: blc(2)             ! bottom left corner
    real(double) :: cache(4,0:256)
    integer                      :: ci(7) = -1         ! colours for grid
    double precision             :: dparam(50)=0.0d0 ! parameters for PGWCSL
    integer                      :: gcode(2) = 1     ! GCODE arg for PGCRVL
    integer                      :: ic
    character(len=80)            :: idents(3)     ! annotation for axes, title
    integer                      :: ierr          
    type(ArrayT)                 :: imageArray
    type(dataSetT)               :: imageSet
    integer                      :: ksl
    integer                      :: labctl       ! label control on pgsbox
    real, parameter              :: tr(6) = (/ 0.0, 1.0, 0.0, 0.0, 0.0, 1.0 /)
    real                         :: trc(2)       ! top right corner
    integer                      :: wcsThingie(4)
    external pgwcsl
    character(len=8) :: ctype(2)
    integer :: minx_temp, miny_temp, i
    real, allocatable :: contours(:)
    real :: pageXMin, pageXMax, pageYMin, pageYMax, pageDeltaX, pageDeltaY,&
      winDeltaX, winDeltaY, pageAspectRatio, winAspectRatio, falseMin, falseMax
    real(double) :: pixelSizeXWorld, pixelSizeYWorld

    !-------------------------------------------------------------------
    ! Start

!    if (withFrame) then
!      imageSet = dataSet(imageSetName, READ)
!      imageArray = array(imageSet, 0, READ)
!      call pgsvp(rhVpXMin, rhVpXMax, rhVpYMin, rhVpYMax)
!if (testFlag) print *, 'in plotImage. minx, maxx, miny, maxy:', minx, maxx, miny, maxy
!      call plotRHinfo(minx, maxx, miny, maxy, imageSet, imageArray)
!      call release(imageSet)
!    end if

    xsize = size(image, 1)
    ysize = size(image, 2)

    if (clipBlankBorder) then
!      call maskArrayLimits(logical(image .ne. 0.0, bool)&
      call maskArrayLimits((image .ne. 0.0)&
        , minx, maxx, miny, maxy, status)
if (status .eq. -1) call error('imageAllZero','There are no non-zero &
&pixels in the image.')
if (status .eq. -2) call error('sizelessImageArray','The image array &
&has zero size.')

    else
      minx = 1
      maxx = xsize
      miny = 1
      maxy = ysize
    end if

    if (withXYclip) then
      minx_temp = nint(minx * (1.0 - xClipFractionLo) + maxx * xClipFractionLo)
      maxx = nint(minx * (1.0 - xClipFractionHi) + maxx * xClipFractionHi)
      minx = minx_temp
      miny_temp = nint(miny * (1.0 - yClipFractionLo) + maxy * yClipFractionLo)
      maxy = nint(miny * (1.0 - yClipFractionHi) + maxy * yClipFractionHi)
      miny = miny_temp
if (minx .ge. maxx) call error('badXclipValues','The requested --xfraclo &
&is greater than or equal to --xfrachi.')
if (miny .ge. maxy) call error('badYclipValues','The requested --yfraclo &
&is greater than or equal to --yfrachi.')
    end if

    if (withFrame) then
imageSet = dataSet(imageSetName, READ)
imageArray = array(imageSet, 0, READ)
call pgsvp(rhVpXMin, rhVpXMax, rhVpYMin, rhVpYMax)
if (testFlag) print *, 'in plotImage. minx, maxx, miny, maxy:', minx, maxx, miny, maxy
call plotRHinfo(minx, maxx, miny, maxy, imageSet, imageArray)
call release(imageSet)

      call pgsvp(mainVpXMin, mainVpXMax, mainVpYMin, mainVpYMax) 
    else
      call pgsvp(0.0, 1.0, 0.0, 1.0) 
    end if
    call pgwnad(minx-0.5, maxx+0.5, miny-0.5, maxy+0.5)

    if (.not. withFrame) then ! set the page size aspect ratio to the same
    ! value as the viewport's. Otherwise have useless black areas.

      call pgqvsz(1, pageXMin, pageXMax, pageYMin, pageYMax)
      pageDeltaX = pageXMax - pageXMin
      pageDeltaY = pageYMax - pageYMin
      pageAspectRatio = pageDeltaY / pageDeltaX
      winDeltaX = maxx - minx
      winDeltaY = maxy - miny
      winAspectRatio = winDeltaY / winDeltaX

      if (pageAspectRatio < winAspectRatio) then ! shrink page delta X:
        pageDeltaX = pageDeltaX * (pageAspectRatio / winAspectRatio)
      end if

      call pgpap(pageDeltaX, winAspectRatio)
    end if

    call pgsch(mainCharSize)

    select case(imageStyle)
    case('contour')
      if (effMin >= effMax) then
        call warning('badBrightnessRange',"The &
        &minimum and maximum image brightness truncation levels are the &
        &same. Can't plot contours.")

      else ! do contours:
        if (contourStyle .eq. 'separation') then
          numContours = 1 + max(0, floor((effMax - effMin) / contourSeparation))
          if (numContours <= 1) call warning('contourSeparationTooLarge'&
            ,'The specified contour separation results in <= 1 contour.')

        else if (contourStyle .ne. 'number') then
call error('badContourStyle','The contour style is not recognized.')
        end if

        allocate(contours(numContours))

        if (numContours == 1) then
          contours(1) = effMin

        else if (contourStyle .eq. 'separation') then
          do i = 1, numContours
            contours(i) = effMin + (i -1) * contourSeparation
          end do

        else ! contourStyle must .eq. 'number':
          do i = 1, numContours
            contours(i) = fracVal(effMin, effMax, (i - 1) / real(numContours))
          end do
        end if

        call pgcons(real(image), xsize, ysize, int(minx), int(maxx)&
          , int(miny), int(maxy), contours, numContours, tr)

        deallocate(contours)
      end if ! effMin >= effMax

    case('image')
      call set_palette(colourMapId)
      call pgsitf(imageTransferId)

      if (effMin >= effMax) then
        call warning('badBrightnessRange',"The &
        &minimum and maximum image brightness truncation levels are the &
        &same.")

        if (effMin == 0) then
          falseMin = -1.0
          falseMax =  1.0
        else if (effMin < 0) then
          falseMin = 1.5 * effMin
          falseMax = 0.5 * effMin
        else
          falseMin = 0.5 * effMin
          falseMax = 1.5 * effMin
        end if

        call pgimag(real(image), xsize, ysize, int(minx), int(maxx)&
          , int(miny), int(maxy), falseMin, falseMax, tr)

      else ! do normal plot:
        call pgimag(real(image), xsize, ysize, int(minx), int(maxx)&
          , int(miny), int(maxy), effMin, effMax, tr)
      end if

    case default
call error('badImageStyle', 'The output image type (ie, colour-coded &
&or contour) is not recognized.')
    end select

    write(errStr, '(a, es10.3)') 'Highest image value shown: ', effMax
    call message(AppMsg, Verbose)
    write(errStr, '(a, es10.3)') 'Lowest image value shown: ', effMin
    call message(AppMsg, Verbose)

    if (withFrame) then
      ! plot title consisting of input file-name at centre of top
      ksl = index(imageSetName, "/", .true.)  ! points to last slash
      call pgmtxt("T", 1.4, 0.5, 0.5, trim(imageSetName(ksl+1:)))

      call pgscf(2)                 ! better quality Roman font

      if(gridStyle .ne. 'none') then
        call pgbox('BC', 0.0, 0, 'BC', 0.0, 0)
        call pglab(' ', wcs%Y%ctype, ' ')
      else
        call pgbox('BCINT', 0.0, 0, 'BCINT', 0.0, 0)
        call pglab('X', 'Y', " ")
      end if
    end if

    ! plot coordinate grid over the top of the image if requested
    if(gridStyle .ne. 'none') then
      idents(1) = wcs%x%ctype
      idents(2) = " "
      idents(3) = " "
      ic        = -1
      dparam(4)     = wcs%x%refPixels 
      dparam(5)     = wcs%y%refPixels 
      dparam(6:9)   = (/1.0d0, 0.0d0, 0.0d0, 1.0d0/)
!      dparam(10) = wcs%X%pixelSizeWorld
!      dparam(11) = wcs%Y%pixelSizeWorld
!vvvvv
! Have to replace the above by the following kludge because pixelSizeWorld is no longer a component of the WcsAxisT type.

    if (wcs%x%ctype .ne. 'RA---TAN' .or. wcs%y%ctype .ne. 'DEC--TAN')&
call error('badWcsType',"Sorry, at present the task will only deal &
&with CTYPE = RA/DEC--TAN.")

    pixelSizeXWorld = -sqrt(wcs%x%matrixOnAxis * wcs%x%matrixOnAxis&
     + wcs%x%matrixOffAxis * wcs%x%matrixOffAxis)
    pixelSizeYWorld = sqrt(wcs%y%matrixOnAxis * wcs%y%matrixOnAxis&
     + wcs%y%matrixOffAxis * wcs%y%matrixOffAxis)
!*** this is only valid if there is no skew.

    select case(wcs%x%worldUnit)
    case('deg')
      dparam(10) = pixelSizeXWorld
    case default
call error('badWcsWorldUnit',"Sorry, at present only degrees are supported.")
    end select

    select case(wcs%y%worldUnit)
    case('deg')
      dparam(11) = pixelSizeYWorld
    case default
call error('badWcsWorldUnit',"Sorry, at present only degrees are supported.")
    end select

!^^^^^

      dparam(20) = wcs%X%refWorld
      dparam(21) = wcs%Y%refWorld
      dparam(22)    = 999d0
      dparam(23)    = 999d0
      dparam(31)    = 0.0d0
      dparam(32) = cos(PiOn180 * wcs%Y%refWorld)&
                 / sin(PiOn180 * wcs%Y%refWorld)
      select case(gridStyle)
      case('ticks')
        gcode = 1
      case('grid')
        gcode = 2
      case default
call error('badGridStyle', 'The grid style is not recognized.')
      end select
      blc(1) = minx - 0.5
      blc(2) = miny - 0.5
      trc(1) = maxx + 0.5
      trc(2) = maxy + 0.5
      if(gridStyle .eq. 'grid') ci(1:2) = 4       ! overlay with blue lines
      labctl = 21

      ctype(1) = wcs%X%ctype
      ctype(2) = wcs%Y%ctype
      call pgsbox(blc, trc, idents, (/"G","E"/), labctl, 0, ci, gcode&
      , -2.0d0, 0, 0.0d0, 0, 0.0d0, .false., pgwcsl, 16, 4, 50, ctype&
      , wcsThingie, dparam, 256, ic, cache, ierr)

      if(ierr /= 0) then
print *, 'pgcrvlErr ierr=', ierr
call error("pgcrvlErr", "error in pgsbox")
      end if
    end if

    if (withFrame .and. imageStyle .eq. 'image') then
      call pgscf(2)

      if (effMin >= effMax) then
        call pgwedg("LI", 4.9, 2.3, falseMin, falseMax, " ")
      else
        call pgwedg("LI", 4.9, 2.3, effMin, effMax, " ")
      end if

      call pgsch(0.5)
      call pgmtxt("T", 1.2, -0.07, 1.0, 'cts/pixel')
      call pgsch(1.0)
      call pgiden
    end if

  end subroutine plotImage

  !-----------------------------------------------------------------------
  subroutine plotRHinfo(minx, maxx, miny, maxy, imageSet, imageArray)
  !-----------------------------------------------------------------------
  ! Prints some info to the right of the image.

    implicit none 

    integer(int16), intent(in) :: minx, maxx, miny, maxy
    type(DataSetT), intent(in) :: imageSet
    type(ArrayT),   intent(in) :: imageArray

    integer :: font, status, slashPos
    character(len=128) :: string
    real :: x, y, dy, csize, xoffset, yoffset
    character(20) :: taskName, versionStr
    character(80) :: legend
    type(TaskInfo) :: info ! type from metatask_mod

    !-------------------------------------------------------------------
    ! Start:

    call pgswin(0.0, 1.0, 0.0, 1.0)

    x = rhCharX
    y = 0.98
    dy = rhCharDy
    xoffset = rhCharXindent
    yoffset = rhCharYindent

    call pgqcf(font)
    call pgscf(1)
    call pgqch(csize)

    info = getTaskInfo() ! gets task name and version number. Routine is in
    ! metatask_mod.

    taskName = nameOf(info)
    versionStr = versionOf(info)
    write(legend, "(a)") trim(taskName)//' v'//trim(versionStr)
    call pgsch(0.7)
    call pgtext(x, y, legend)

    y = y - 2.0 * dy
    call pgsch(rhCharSize)

    call pgtext(x, y, 'Instrument:')
    call getStringKwd('INSTRUME', imageSet, imageArray, string, status)
    if(status == 0) call pgtext(x+xoffset, y+yoffset, trim(string))

    call getStringKwd('FILTER', imageSet, imageArray, string, status)
    y = y - dy
    call pgtext(x, y, 'Filter:')
    if(status == 0) call pgtext(x+xoffset, y+yoffset, trim(string))

    call getStringKwd('OBJECT', imageSet, imageArray, string, status)
    y = y - dy
    call pgtext(x, y, 'Object:')
    if(status == 0) call pgtext(x+xoffset, y+yoffset, trim(string))

    call getStringKwd('OBSERVER', imageSet, imageArray, string, status)
    y = y - dy
    call pgtext(x, y, 'Observer:')
    if(status == 0) call pgtext(x+xoffset, y+yoffset, trim(string))

    call getStringKwd('DATE-OBS', imageSet, imageArray, string, status)
    y = y - dy
    call pgtext(x, y, 'DATE-OBS:')
    if(status == 0) call pgtext(x+xoffset, y+yoffset, string(1:16))

    call getStringKwd('EXPOSURE', imageSet, imageArray, string, status)
    y = y - dy
    call pgtext(x, y, 'Exposure:')
    if(len_trim(string) > 0 .and. status == 0) then
      string = trim(string) // ' sec'
    call pgtext(x+xoffset, y+yoffset, trim(string)) 
    end if

if (testFlag) print *, 'in plotRHinfo. minx, maxx, miny, maxy:', minx, maxx, miny, maxy
    write(string, "('pixels (', i0, ':', i0, ',', i0, ':', i0, ')')") &
      minx, maxx, miny, maxy
    y = y - 1.5 * dy
    call pgtext(x, y, 'Image size:')
    call pgtext(x+xoffset, y+yoffset, trim(string))


    y = y - 1.5 * dy
    call pgtext(x, y, 'Parameters:')
    y = y - 0.5 * dy
    call pgtext(x+xoffset, y, 'scaletype = '//trim(imageTransferStyle))
    y = y - 0.5 * dy
    write(string, "('zlo = ', f8.6)") zClipFractionLo
    call pgtext(x+xoffset, y, trim(string))
    y = y - 0.5 * dy
    write(string, "('zhi = ', f8.6)") zClipFractionHi
    call pgtext(x+xoffset, y, trim(string))
    if (withSrcList) then
      y = y - 0.5 * dy
      write(string, "('radius = ', f6.2)") maxRadiusPixels
      call pgtext(x+xoffset, y, trim(string))
    end if
    if (withXYclip) then
      y = y - 0.5 * dy
      write(string, "('xlo = ', f6.4)") xClipFractionLo
      call pgtext(x+xoffset, y, trim(string))
      y = y - 0.5 * dy
      write(string, "('xhi = ', f6.4)") xClipFractionHi
      call pgtext(x+xoffset, y, trim(string))
      y = y - 0.5 * dy
      write(string, "('ylo = ', f6.4)") yClipFractionLo
      call pgtext(x+xoffset, y, trim(string))
      y = y - 0.5 * dy
      write(string, "('yhi = ', f6.4)") yClipFractionHi
      call pgtext(x+xoffset, y, trim(string))
    end if
    if (addUserSources) then
      if (size(userRaDeg) .le. 1) then
        y = y - 0.5 * dy
        write(string, "('user RA (deg) = ', f10.5)") userRaDeg(1)
        call pgtext(x+xoffset, y, trim(string))
        y = y - 0.5 * dy
        write(string, "('user dec (deg) = ', f10.5)") userDecDeg(1)
        call pgtext(x+xoffset, y, trim(string))
      else
        y = y - dy
        string = intToStr(size(userRaDeg))
        string = trim(string)//' user sources added.'
        call pgtext(x, y, trim(string))
      end if
    end if
    y = y - 0.5 * dy
    write(string, "('colourmap = ', i3)") colourMapId
    call pgtext(x+xoffset, y, trim(string))



    if (withSrcList) then
      ! Annotate plot with source list filename:

      y = 0.01

      call pgtext(x, y, "Sources from:")
      slashPos = index(srcListSetTabName, "/", .true.)
      call pgtext(x+xoffset, y+yoffset, trim(srcListSetTabName(slashPos+1:)))
    end if


    call pgscf(font)
    call pgsch(csize)

  end subroutine plotRHinfo

  !-----------------------------------------------------------------------
  subroutine plotRHinfo_20030119(minx, maxx, miny, maxy, imageSet, imageArray)
  !-----------------------------------------------------------------------
  ! shows FITS headers at top of plot

!    use fitsutils_mod ! contains getStringKwd() etc and intToStr()

    implicit none 

!    integer,        intent(in) :: minx, maxx, miny, maxy
integer(int16),        intent(in) :: minx, maxx, miny, maxy
    ! world coordinate limits
    type(DataSetT), intent(in) :: imageSet
    type(ArrayT),   intent(in) :: imageArray

    integer :: font, status
    character(len=128) :: string
    real :: x, y, dy, csize, xoffset, yoffset

    !-------------------------------------------------------------------
    ! Start:

    x = rhCharX
    y = 0.95
    dy = rhCharDy
xoffset = rhCharXindent
yoffset = rhCharYindent

    call pgqcf(font)
    call pgscf(1)
    call pgqch(csize)
!    call pgsch(rhTextSize)
call pgsch(rhCharSize)

    call pgmtxt("RV", x, y, 0.0, 'Instrument:')
    call getStringKwd('INSTRUME', imageSet, imageArray, string, status)
    if(status .eq. 0) call pgmtxt("RV", x+xoffset, y+yoffset, 0.0, trim(string))

    call getStringKwd('FILTER', imageSet, imageArray, string, status)
    y = y - dy
    call pgmtxt("RV", x, y, 0.0, 'Filter:')
    if(status .eq. 0) call pgmtxt("RV", x+xoffset, y+yoffset, 0.0, trim(string))

    call getStringKwd('OBJECT', imageSet, imageArray, string, status)
    y = y - dy
    call pgmtxt("RV", x, y, 0.0, 'Object:')
    if(status .eq. 0) call pgmtxt("RV", x+xoffset, y+yoffset, 0.0, trim(string))

    call getStringKwd('OBSERVER', imageSet, imageArray, string, status)
    y = y - dy
    call pgmtxt("RV", x, y, 0.0, 'Observer:')
    if(status .eq. 0) call pgmtxt("RV", x+xoffset, y+yoffset, 0.0, trim(string))

    call getStringKwd('DATE-OBS', imageSet, imageArray, string, status)
    y = y - dy
    call pgmtxt("RV", x, y, 0.0, 'DATE-OBS:')
    if(status .eq. 0) call pgmtxt("RV", x+xoffset, y+yoffset, 0.0, string(1:16))

    call getStringKwd('EXPOSURE', imageSet, imageArray, string, status)
    y = y - dy
    call pgmtxt("RV", x, y, 0.0, 'Exposure:')
    if(len_trim(string) > 0 .and. status .eq. 0) then
      string = trim(string) // ' sec'
      call pgmtxt("RV", x+xoffset, y+yoffset, 0.0, trim(string)) 
    end if


    write(string, "('pixels (', i0, ':', i0, ',', i0, ':', i0, ')')") &
      minx, maxx, miny, maxy
    y = y - 1.5 * dy
    call pgmtxt("RV", x, y, 0.0, 'Image size:')
    call pgmtxt("RV", x+xoffset, y+yoffset, 0.0, trim(string))


    y = y - 1.5 * dy
    call pgmtxt("RV", x, y, 0.0, 'Parameters:')
    y = y - 0.5 * dy
    call pgmtxt("RV", x+xoffset, y, 0.0, 'scaletype = '&
    //trim(imageTransferStyle))
    y = y - 0.5 * dy
    write(string, "('zlo = ', f8.6)") zClipFractionLo
    call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
    y = y - 0.5 * dy
    write(string, "('zhi = ', f8.6)") zClipFractionHi
    call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
    if (withSrcList) then
      y = y - 0.5 * dy
!      write(string, "('radius = ', f6.2)") userRadiusPixels
      write(string, "('radius = ', f6.2)") maxRadiusPixels
      call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
    end if
    if (withXYclip) then
      y = y - 0.5 * dy
      write(string, "('xlo = ', f6.4)") xClipFractionLo
      call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
      y = y - 0.5 * dy
      write(string, "('xhi = ', f6.4)") xClipFractionHi
      call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
      y = y - 0.5 * dy
      write(string, "('ylo = ', f6.4)") yClipFractionLo
      call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
      y = y - 0.5 * dy
      write(string, "('yhi = ', f6.4)") yClipFractionHi
      call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
    end if
    if (addUserSources) then
!      if (size(userSrcData) .le. 1) then
      if (size(userRaDeg) .le. 1) then
        y = y - 0.5 * dy
!        write(string, "('user RA (deg) = ', f10.5)") userSrcData(1)%raDeg
        write(string, "('user RA (deg) = ', f10.5)") userRaDeg(1)
        call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
        y = y - 0.5 * dy
!        write(string, "('user dec (deg) = ', f10.5)") userSrcData(1)%decDeg
        write(string, "('user dec (deg) = ', f10.5)") userDecDeg(1)
        call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
      else
        y = y - dy
!        string = intToStr(size(userSrcData))
        string = intToStr(size(userRaDeg))
        string = trim(string)//' user sources added.'
        call pgmtxt("RV", x, y, 0.0, trim(string))
      end if
    end if
    y = y - 0.5 * dy
    write(string, "('colourmap = ', i3)") colourMapId
    call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))

    call pgscf(font)
    call pgsch(csize)

  end subroutine plotRHinfo_20030119

  !----------------------------------------------------------------------
  subroutine addSources(minx, maxx, miny, maxy)
  !----------------------------------------------------------------------

    implicit none

    integer(int16), intent(in) :: minx, maxx, miny, maxy ! image pixel ranges

    character(20) ::&
      radiusColName,&
      sortColName
    type(dataSetT) ::&
      sourceSet, imageSet
    type(TableT) ::&
      sourceTab
    integer(int32) ::&
      numFilteredRows,&
      numLastSource,&
      src,&
      numOutside,&
      numShown
    real, pointer ::&
      xAmp(:), yAmp(:), phase(:), srcXpixels(:), srcYpixels(:)
!    logical(bool), pointer :: srcIsValid(:)
    logical, pointer :: srcIsValid(:)
    real, allocatable ::&
      tempRadii(:)
    integer(int32), pointer ::&
      indices(:),&
      tempIndices(:)
    type(ColumnT) ::&
      radiusCol
    real(single) ::&
      maxRadiusDataValue
    real(double), allocatable ::&
      sortColData(:),&
      radiusColData(:),&
      raDeg(:),&
      decDeg(:),&
      srcEeV(:)
    real(double) ::&
      cel(10),&
      prj(0:20),&
      lin(20),&
      world(2),&
      imgcrd(2),&
      pixcrd(2),&
      phi,&
      theta
    real ::&
      x,&
      y,&
      csize,&
      minRadiusFactor = 0.20,&
      maxRadiusFactor = 1.0
    integer ::&
      wcsThingie(0:3),&
      colsave,&
      fontsave,&
      thicksave,&
      j,&
      ierr
!    logical(bool) ::&
    logical ::&
      allRowsNull
    real(double) :: crval(2), pixelSizeXWorld, pixelSizeYWorld
    character(8) :: ctype(2)
    character(20), allocatable ::&
      labelStrings(:)
integer :: i
  type(WcsT) :: srcWcs
logical :: testFlag = .false.
    real(single), allocatable :: srcCounts(:), srcBkgCtsPerPixel(:)

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, 'Open source list.')

    !....................................................................
    ! Load and filter source list:

    call splitSetTabName(srcListSetTabName, srcListSetName, srcListTabName)

    sourceSet = clone(srcListSetName, 'tempset.ds', TEMP)
!sourceSet = clone(srcListSetName, tempSrcSetName, TEMP)
!*** get rid of this parameter, it is not needed after all.

    sourceTab = table(sourceSet, srcListTabName, MODIFY)

    call message(AppMsg, Verbose, 'Selection of sources.')

    call initFilter(selectionExpression)
    numFilteredRows = doFilter(sourceTab)

    if (numFilteredRows <= 0)&
      call warning('noSourcesSelected'&
        , 'No sources passed the specified filtering expression.')

    !....................................................................
    ! Calculate radii of source circles:

    call message(AppMsg, Verbose, 'Calculating radii.')

    allocate(raDeg(numFilteredRows), decDeg(numFilteredRows))
    call readColDataToFixed(sourceTab, 'RA',  raDeg)
    call readColDataToFixed(sourceTab, 'DEC', decDeg)

    select case(radiusStyle)
    case('user')
      allocate(xAmp(numFilteredRows), yAmp(numFilteredRows)&
        , phase(numFilteredRows), srcIsValid(numFilteredRows)&
        , srcXpixels(numFilteredRows), srcYpixels(numFilteredRows))

      yAmp = userRadiusPixels
      xAmp = yAmp

    case('expr')
      allocate(xAmp(numFilteredRows), yAmp(numFilteredRows)&
        , phase(numFilteredRows), srcIsValid(numFilteredRows)&
        , srcXpixels(numFilteredRows), srcYpixels(numFilteredRows))

      phase = PiOverTwo ! default

      radiusColName = radiusExpression
      radiusCol = column(sourceTab, radiusColName)

      call maxNonNullValue(sourceTab, radiusColName, maxRadiusDataValue&
        , allRowsNull) ! in ssclib/dal_aux

      if (allRowsNull)&
call error('allRadiusRowsNull', 'All the rows specified by &
&--radiusexpression are null-valued.')

      allocate(radiusColData(numFilteredRows))
      call readColDataToFixed(radiusCol, radiusColData) ! in ssclib/dal_aux
      yAmp = 1.0 + 0.5 * log10(radiusColData/maxRadiusDataValue)
      deallocate(radiusColData)
      where(yAmp .gt. maxRadiusFactor) yAmp = maxRadiusFactor
      where(yAmp .lt. minRadiusFactor) yAmp = minRadiusFactor
      yAmp = maxRadiusPixels * yAmp

      xAmp = yAmp
      srcIsValid = .true.

    case('psf')
    imageSet = dataSet(imageSetName, READ)
    call getCalInfo(imageSet, instrumentId, expStartTimeStamp, scAttitude&
      , instrumAttitude) ! sets CAL instrument state.

    if (instrumentId .ne. EMOS1&
    .and. instrumentId .ne. EMOS2&
    .and. instrumentId .ne. EPN) then
call error('badInstrum', 'Instrument '//trim(instId(instrumentId))//&
&' not supported.')
    end if

    call release(imageSet)

!srcWcs%X%refPixels       = int32Attribute(sourceSet, 'TCRPIX1')
!srcWcs%Y%refPixels       = int32Attribute(sourceSet, 'TCRPIX2')
!srcWcs%X%pixelSizeWorld = real32Attribute(sourceSet, 'TCDELT1')
!srcWcs%Y%pixelSizeWorld = real32Attribute(sourceSet, 'TCDELT2')
!srcWcs%X%refWorld       = real32Attribute(sourceSet, 'TCRVAL1')
!srcWcs%Y%refWorld       = real32Attribute(sourceSet, 'TCRVAL2')
!srcWcs%X%ctype          = stringAttribute(sourceSet, 'TCTYPE1')
!srcWcs%Y%ctype          = stringAttribute(sourceSet, 'TCTYPE2')

call readWcsSourceList(sourceSet, srcWcs)
!*** they should be stored as proper column keywords, but until Potsdam change their code there's not much I can do.


!call getSourceCutouts_20031029(sourceTab, wcs, srcWcs, scAttitude&
!, expStartTimeStamp, fracBackground, 'RA', 'DEC'&
!, 'SCTS', 'BG_MAP', srcXpixels, srcYpixels&
!, xAmp, yAmp, phase, srcIsValid)

! replaced by: vvvvv
!allocate(srcRaDeg(numSources), srcDecDeg(numSources))
allocate(srcCounts(numFilteredRows), srcBkgCtsPerPixel(numFilteredRows))
allocate(srcEeV(numFilteredRows))

!call readColDataToFixed(sourceTab, 'RA',      srcRaDeg)
!call readColDataToFixed(sourceTab, 'DEC',     srcDecDeg)
call readColDataToFixed(sourceTab, 'SCTS',    srcCounts)
call readColDataToFixed(sourceTab, 'BG_MAP',  srcBkgCtsPerPixel)

if (all(srcBkgCtsPerPixel <= 0.0)) then
  srcBkgCtsPerPixel = 0.1
else
  srcBkgCtsPerPixel = median(srcBkgCtsPerPixel) ! in ssclib/array_utils
end if

srcEeV = 1000.0d0

allocate(xAmp(numFilteredRows), yAmp(numFilteredRows)&
, phase(numFilteredRows), srcIsValid(numFilteredRows), srcXpixels(numFilteredRows), srcYpixels(numFilteredRows))

call getSourceCutouts(wcs, srcWcs, scAttitude&
, expStartTimeStamp, fracBackground, srcEeV, raDeg, decDeg&
, srcCounts, srcBkgCtsPerPixel, srcXpixels, srcYpixels&
, xAmp, yAmp, phase, srcIsValid) ! in ssclib/source_cutouts
!^^^^^^^^^^^^^^^^^^^^

if (testFlag) then
print *, '** now back out in implot.'
do src = 1, size(raDeg)
  print *, 'src, xAmp(src), yAmp(src), phase(src): '&
  , src, xAmp(src), yAmp(src), phase(src)
end do
print *, ''
end if

!*** srcXpixels, srcYpixels not used at present.

    end select ! radiusStyle

    !....................................................................
    ! If a limit on the number of sources to be plotted is desired, the sources
    ! must first be sorted. This is accomplished next:

    allocate(indices(numFilteredRows)&
    , tempRadii(numFilteredRows))

    if (withMaxSrcNum) then
      call message(AppMsg, Verbose, 'Sorting to prepare for max source &
      &number cut.')

      select case(maxSourceNumSortStyle)
      case('radiusup')
        tempRadii = yAmp
        allocate(tempIndices(numFilteredRows))
        tempIndices = (/(i, i=1,numFilteredRows)/)
        call quickSort(tempRadii, tempIndices)
        do i = 1, size(tempIndices)
          indices(i) = tempIndices(i)
        end do
        deallocate(tempIndices)

      case('radiusdown')
        tempRadii = -yAmp
        allocate(tempIndices(numFilteredRows))
        tempIndices = (/(i, i=1,numFilteredRows)/)
        call quickSort(tempRadii, tempIndices)
        do i = 1, size(tempIndices)
          indices(i) = tempIndices(i)
        end do
        deallocate(tempIndices)
  
      case('expr')
        sortColName = maxSourceNumSortExpression

        allocate(sortColData(numFilteredRows))
        call readColDataToFixed(sourceTab, sortColName, sortColData)
        sortColData = -sortColData
        allocate(tempIndices(numFilteredRows))
        tempIndices = (/(i, i=1,numFilteredRows)/)
        call quickSort(sortColData, tempIndices)
        do i = 1, size(tempIndices)
          indices(i) = tempIndices(i)
        end do
        deallocate(sortColData)
        deallocate(tempIndices)

      case('rownumber')
        allocate(tempIndices(numFilteredRows))
        tempIndices = (/(i, i=1,numFilteredRows)/)
        indices = tempIndices
        deallocate(tempIndices)

      case default
call error('badNcutSortStyle', 'The value of --ncutsortstyle is &
&not recognized.')
      end select

    else ! no withMaxSrcNum given.
      allocate(tempIndices(numFilteredRows))
      tempIndices = (/(i, i=1,numFilteredRows)/)
      indices = tempIndices
      deallocate(tempIndices)
    end if ! withMaxSrcNum

    !....................................................................
    ! Allocating source labels:

    if (withSrcLabels) then
      call message(AppMsg, Verbose, 'Allocating source labels.')

      allocate(labelStrings(numFilteredRows))
      call getLabelStrings(sourceTab, yAmp, labelStrings)
    end if

    deallocate(tempRadii)

    !....................................................................
    ! Preparing for plot:

    wcsThingie = (/ 0, 1, 2, 0/)
    cel(1:4)  = (/ wcs%X%refWorld, wcs%Y%refWorld, 999.0d0, 999.0d0 /)
    cel(5:10) = 0.0d0
    prj  =  0.0d0
!    lin(1:11) = (/ 0.0d0, 2.0d0, 2.0d0, dble(wcs%X%refPixels)&
!    , dble(wcs%Y%refPixels)&
!    , 1.0d0, 0.0d0, 0.0d0, 1.0d0, wcs%X%pixelSizeWorld, wcs%Y%pixelSizeWorld /)

!vvvvv
! Have to replace the above by the following kludge because pixelSizeWorld is no longer a component of the WcsAxisT type.

    if (wcs%x%ctype .ne. 'RA---TAN' .or. wcs%y%ctype .ne. 'DEC--TAN')&
call error('badWcsType',"Sorry, at present the task will only deal &
&with CTYPE = RA/DEC--TAN.")

    if (wcs%x%worldUnit .ne. 'deg' .or. wcs%y%worldUnit .ne. 'deg')&
call error('badWcsWorldUnit',"Sorry, at present only degrees are supported.")

    pixelSizeXWorld = -sqrt(wcs%x%matrixOnAxis * wcs%x%matrixOnAxis&
     + wcs%x%matrixOffAxis * wcs%x%matrixOffAxis)
    pixelSizeYWorld = sqrt(wcs%y%matrixOnAxis * wcs%y%matrixOnAxis&
     + wcs%y%matrixOffAxis * wcs%y%matrixOffAxis)
!*** this is only valid if there is no skew.

lin(1:11) = (/ 0.0d0, 2.0d0, 2.0d0, dble(wcs%X%refPixels)&
 , dble(wcs%Y%refPixels)&
 , 1.0d0, 0.0d0, 0.0d0, 1.0d0, pixelSizeXWorld, pixelSizeYWorld /)
!^^^^^

    crval(1) = wcs%X%refWorld
    crval(2) = wcs%Y%refWorld
    ctype(1) = wcs%X%ctype
    ctype(2) = wcs%Y%ctype

    call pgqci(colsave)
    call pgsci(textClrIndex)
    call pgqcf(fontsave)
    call pgscf(1)
    call pgsfs(2)    ! for hollow circles
    call pgqlw(thicksave)
    call pgslw(lineThicknessPixels)
    call pgqch(csize)
    call pgsch(textSize)

    !....................................................................
    ! Plot the source circles + labels:

    call message(AppMsg, Verbose, 'Plotting sources.')

    if (withMaxSrcNum) then
      numLastSource = min(userMaxSourceNum, numFilteredRows)
    else
      numLastSource = numFilteredRows
    end if

    numOutside = 0
    numShown = 0
    do src = 1, numLastSource
      j = indices(src)

      ! transfer ra/dec to wcs
!      world(1) = raDeg(j-1)
!      world(2) = decDeg(j-1)
world(1) = raDeg(j)
world(2) = decDeg(j)
      call wcsfwd(ctype, wcsThingie, world, crval, cel, phi, theta, prj, imgcrd&
      , lin, pixcrd, ierr)

      x = real(pixcrd(1)) 
      y = real(pixcrd(2)) 
x = x + 0.5
y = y + 0.5
!**** I suspect that there is some such error in source plots.

      if(x < minx .or. x > maxx .or. y < miny .or. y > maxy) then
        numOutside = numOutside + 1

      else
        numShown = numShown + 1

        if (radiusStyle .eq. 'psf' .and. withEllipse) then
          call drawSrcEllipse(x, y, xAmp(j), yAmp(j), phase(j))
if (testFlag) then
  print *, 'j, x, y, xAmp(j), yAmp(j), phase(j): '&
  , j, x, y, xAmp(j), yAmp(j), phase(j)
end if

        else ! draw hollow circle:
          call pgcirc(x, y, yAmp(j))
        end if

        if(withSrcLabels) then
          call pgptxt(x + yAmp(j) + 1.0, y, 0.0, 0.0&
            , trim(labelStrings(j)))
        end if
      end if
    end do

    if(numOutside > 0) then
      write(errStr, '(i6,1x,a)') numOutside, 'source positions outside image.'
      call warning('sourcesOutside')
    end if

    if (addUserSources) then
      do src = 1, size(userRaDeg)
!        world(1) = userSrcData(src)%raDeg
!        world(2) = userSrcData(src)%decDeg
world(1) = userRaDeg(src)
world(2) = userDecDeg(src)
        call wcsfwd(ctype, wcsThingie, world, crval, cel, phi, theta&
        , prj, imgcrd, lin, pixcrd, ierr)

        x = real(pixcrd(1)) 
        y = real(pixcrd(2))
        call pgsci(userSrcClrIndices(src))
        call pgpt1(x, y, 2)
        call pgptxt(x + 3.0, y + 2.0, 0.0, 0.0, trim(userLabels(src)))
      end do
    end if

    !....................................................................
    ! Add notations to base and side of plot:

    if (withFrame) then
      call message(AppMsg, Verbose, 'Remaining inscriptions.')

      call pgsvp(botVpXMin, botVpXMax, botVpYMin, botVpYMax)
      call plotBottomInfo(numLastSource, radiusColName)
    end if ! withFrame

    !....................................................................
    call message(AppMsg, Verbose, 'Clean up.')

    call pgsci(colsave)
    call pgscf(fontsave)
    call pgsch(csize)

    deallocate(indices)
    deallocate(xAmp, yAmp, phase, srcIsValid, srcXpixels, srcYpixels)
    deallocate(raDeg)
    deallocate(decDeg)

    call message(AppMsg, Verbose, 'Release source set.')

    call release(sourceSet)

  end subroutine addSources

  !---------------------------------------------------------------------
  subroutine drawSrcEllipse(x, y, xAmp, yAmp, phase)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: x, y, xAmp, yAmp, phase

    real :: xVals(20), yVals(20)

    !-------------------------------------------------------------------
    ! Start

    if (xAmp .gt. 0.0 .and. yAmp .gt. 0.0) then
      call calcEllipseFromPhase(xAmp, yAmp, phase, xVals, yVals)
      xVals = xVals + x
      yVals = yVals + y
      call pgpoly(size(xVals), xVals, yVals)
    else
      call pgpt1(x, y, 2)
    end if

  end subroutine drawSrcEllipse

  !---------------------------------------------------------------------
  subroutine plotBottomInfo(numLastSource, radiusColName)
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in) :: numLastSource
    character(*),   intent(in) :: radiusColName

    real :: x, y, dy
    character(255) ::&
      text

    !-------------------------------------------------------------------
    ! Start

    call pgswin(0.0, 1.0, 0.0, 1.0)
    call pgsci(1)
    call pgsch(botCharSize)

    x = botCharX
    y = 0.8
    dy = botCharDy
    call pgtext(x, y, "Selection expr: '"//trim(selectionExpression)//"'")


    if (withMaxSrcNum) then
      select case(maxSourceNumSortStyle)
      case('radiusup')
        write(text, "(a, i0, a)") "Number of sources: first ", numLastSource&
        , " in increasing order of "//trim(radiusExpression)//"."

      case('radiusdown')
        write(text, "(a, i0, a)") "Number of sources: first ", numLastSource&
        , " in decreasing order of "//trim(radiusExpression)//"."

      case('expr')
        write(text, "(a, i0, a)") "Number of sources: first ", numLastSource&
        , " in decreasing order of "//trim(maxSourceNumSortExpression)//"."

      case('rownumber')
        write(text, "(a, i0, a)") "Number of sources: first ", numLastSource&
        , " in increasing order of row number."
      end select
    else
      write(text, "(a, i0, a)") "Number of sources: all "&
      , numLastSource, " selected."
    end if

    y = y - dy
    call pgtext(x, y, text)

    if (radiusStyle .eq. 'expr') then
      text = "Radii: scaled from expression "//trim(radiusExpression)//"."
    else if (radiusStyle .eq. 'column') then
      text = "Radii: scaled from column "//trim(radiusColName)//"."
    else if (radiusStyle .eq. 'psf') then
      text = "Radii: scaled from PSF spread."
    else
      text = "Radii: fixed."
    end if

    y = y - dy
    call pgtext(x, y, text)

    if (withSrcLabels) then
      select case(srcLabelStyle)
      case('radiusup')
        text = 'Labels: integer, in order of increasing radius.'
      case('radiusdown')
        text = 'Labels: integer, in order of decreasing radius.'
      case('expr')
        text = 'Labels: from column '//trim(srcLabelExpression)//'.'
      case('rownumber')
        text = 'Labels: from source list row numbers.'
      end select
      
      y = y - dy
      call pgtext(x, y, text)
    end if ! withSrcLabels

  end subroutine plotBottomInfo

  !---------------------------------------------------------------------
  subroutine getLabelStrings(sourceTab, yAmp, labelStrings)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)  :: sourceTab
    real, intent(in)  :: yAmp(:)
    character(*), intent(out) :: labelStrings(size(yAmp))

    integer(int32) :: numFilteredRows, i
    real(single), allocatable ::&
      tempRadii(:)
    integer(int32), pointer :: indices(:)
    integer(int32), allocatable :: integerLabels(:)
    real(double), allocatable :: realLabels(:)
    character(20) :: labelsColName
    type(ColumnT) :: col
    logical(bool), pointer :: booleanData(:)

    !-------------------------------------------------------------------
    ! Start

    numFilteredRows = size(yAmp)

    allocate(tempRadii(numFilteredRows))

    select case(srcLabelStyle)
    case('radiusup', 'radiusdown')
      tempRadii = yAmp ! can't alter yAmp as intent(in).
      if (srcLabelStyle .eq. 'radiusdown') tempRadii = -tempRadii

      allocate(indices(numFilteredRows))
      indices = (/(i, i=1,numFilteredRows)/)
      call quickSort(tempRadii, indices)
      do i = 1, numFilteredRows
        labelStrings(indices(i)) = intToStr(i, leftJustify=.true._boolean)
      end do
      deallocate(indices)

    case('expr')
      labelsColName = srcLabelExpression

      select case(dataType(column(sourceTab, labelsColName)))
      case(INTEGER8, INTEGER16, INTEGER32)
        allocate(integerLabels(numFilteredRows))
        call readColDataToFixed(sourceTab, labelsColName, integerLabels)
        do i = 1, numFilteredRows
          labelStrings(i) = intToStr(integerLabels(i)&
            , leftJustify=.true._boolean)
        end do
        deallocate(integerLabels)

      case(REAL32, REAL64)
        allocate(realLabels(numFilteredRows))
        call readColDataToFixed(sourceTab, labelsColName, realLabels)
        do i = 1, numFilteredRows
          write(labelStrings(i), '(es12.5)') realLabels(i)
        end do
        deallocate(realLabels)

      case(STRING)
        col = column(sourceTab, labelsColName)
        do i = 1, numFilteredRows
          labelStrings(i) = ''
          labelStrings(i) = stringCell(col, i-1)
        end do
        call release(col)

      case(BOOLEAN)
        col = column(sourceTab, labelsColName)
        booleanData => boolData(col)
        do i = 1, numFilteredRows
          if (booleanData(i-1)) then
            labelStrings(i) = 'T'
          else
            labelStrings(i) = 'F'
          end if
        end do
        call release(col)

      case default
call error('badLabelColumnDataType',"The data type of the label &
&column"//trim(labelsColName)//" is not recognized.")
      end select

    case('rownumber')
      allocate(indices(numFilteredRows))
      indices = (/(i, i=1,numFilteredRows)/)
      do i = 1, numFilteredRows
        labelStrings(i) = intToStr(indices(i), leftJustify=.true._boolean)
      end do
      deallocate(indices)

    case default
call error('badLabelsStyle', 'The label style is not recognized.')
    end select

  end subroutine getLabelStrings

!-----------------------------------------------------------------------
! set_palette: Sets a color palette for use by PGIMAG
!-----------------------------------------------------------------------
subroutine set_palette(itype)
integer, intent(in) :: itype    !colour palette: -1=inverse gray, 1=grey, 2=rainbow, 3=heat, 
                                ! 4=IRAF, 5=AIPS, 6=TJP
real, parameter :: al(20) = (/0.0, 0.1, 0.1, 0.2, 0.2, 0.3, 0.3, 0.4, 0.4, 0.5, &
                              0.5, 0.6, 0.6, 0.7, 0.7, 0.8, 0.8, 0.9, 0.9, 1.0/)
real, parameter :: ar(20) = (/0.0, 0.0, 0.3, 0.3, 0.5, 0.5, 0.0, 0.0, 0.0, 0.0, &
                              0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0/)
real, parameter :: ag(20) = (/0.0, 0.0, 0.3, 0.3, 0.0, 0.0, 0.0, 0.0, 0.8, 0.8, &
                              0.6, 0.6, 1.0, 1.0, 1.0, 1.0, 0.8, 0.8, 0.0, 0.0/)
real, parameter :: ab(20) = (/0.0, 0.0, 0.3, 0.3, 0.7, 0.7, 0.7, 0.7, 0.9, 0.9, &
                              0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0/)
real, parameter :: grey(2) = (/0.0, 1.0/)
real, parameter :: invg(2) = (/1.0, 0.0/)
real, parameter :: hl(5) = (/0.0, 0.2, 0.4, 0.6, 1.0/)
real, parameter :: hr(5) = (/0.0, 0.5, 1.0, 1.0, 1.0/)
real, parameter :: hg(5) = (/0.0, 0.0, 0.5, 1.0, 1.0/)
real, parameter :: hb(5) = (/0.0, 0.0, 0.0, 0.3, 1.0/)
real, parameter :: rl(9) = (/-0.5, 0.0, 0.17, 0.33, 0.50, 0.67, 0.83, 1.0, 1.7/)
real, parameter :: rr(9) = (/0.0, 0.0, 0.0, 0.0, 0.6, 1.0, 1.0, 1.0, 1.0/)
real, parameter :: rg(9) = (/0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 0.6, 0.0, 1.0/)
real, parameter :: rb(9) = (/0.0, 0.3, 0.8, 1.0, 0.3, 0.0, 0.0, 0.0, 1.0/)
real, parameter :: tl(4) = (/0.0, 0.5, 0.5, 1.0/)
real, parameter :: tr(4) = (/0.2, 0.6, 0.6, 1.0/)
real, parameter :: tg(4) = (/0.0, 0.0, 0.5, 1.0/)
real, parameter :: tb(4) = (/1.0, 0.0, 0.0, 0.0/)
real, parameter :: wl(10) = (/0.0, 0.5, 0.5, 0.7, 0.7, 0.85, 0.85, 0.95, &
                              0.95, 1.0/)
real, parameter :: wr(10) = (/0.0, 1.0, 0.0, 0.0, 0.3,  0.8,  0.3,  1.0, &
                              1.0, 1.0/)
real, parameter :: wg(10) = (/0.0, 0.5, 0.4, 1.0, 0.0,  0.0,  0.2,  0.7,  &
                              1.0, 1.0/)
real, parameter :: wb(10) = (/0.0, 0.0, 0.0, 0.0, 0.4,  1.0,  0.0,  0.0, &
                              0.95, 1.0/)
! following designed by Clive Page: black + rainbow + white at top end
real, parameter :: cl(7) = (/0.1,  0.25, 0.43, 0.57, 0.71, 0.8,  1.0/)
real, parameter :: cr(7) = (/0.5,  0.0,  0.0,  0.0,  0.9,  1.0,  1.0/)
real, parameter :: cg(7) = (/0.0,  0.0,  0.8,  1.0,  0.9,  0.3,  1.0/)
real, parameter :: cb(7) = (/0.5,  1.0,  0.8,  0.0,  0.0,  0.3,  1.0/)
!                          black  mag   blue  cyan  grn   yell  red   white

select case(itype)
case(-2)                                             ! inverse rainbow
   call pgctab(rl, rr(9:1:-1), rg(9:1:-1), rb(9:1:-1), 9, 1.0, 0.5)
case(-1) 
   call pgctab(grey, invg, invg, invg, 2, 1.0, 0.5)  ! inverse greyscale
case(0)
call error('badColourMap', 'Sorry, --colourmap=0 is not allowed.')
case(1)
   call pgctab(grey, grey, grey, grey, 2, 1.0, 0.5)  ! greyscale
case(2)
   call pgctab(rl, rr, rg, rb, 9, 1.0, 0.5)          ! rainbow
case(3)
   call pgctab(hl, hr, hg, hb, 5, 1.0, 0.5)          ! heat
case(4)
   call pgctab(wl, wr, wg, wb, 10, 1.0, 0.5)         ! weird IRAF
case(5)
   call pgctab(al, ar, ag, ab, 20, 1.0, 0.5)         ! AIPS
case(6)
   call pgctab(tl, tr, tg, tb, 4, 1.0, 0.5)          ! Tim Pearson
case(7)
   call pgctab(cl, cr, cg, cb, 8, 1.0, 0.5)         ! Clive Page
   call pgscr(16, 0.0, 0.0, 0.0)                    ! bottom level is black
end select
return
end subroutine set_palette

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module implot_mod
